from monocle_apptrace.instrumentation.common.constants import SPAN_SUBTYPES, SPAN_TYPES
from monocle_apptrace.instrumentation.common.utils import get_error_message
from monocle_apptrace.instrumentation.metamodel.adk import _helper
TOOL = {
      "type": SPAN_TYPES.AGENTIC_TOOL_INVOCATION,
      "subtype": SPAN_SUBTYPES.ROUTING,
      "attributes": [
        [
              {
                "_comment": "tool type",
                "attribute": "type",
                "accessor": lambda arguments:'tool.adk'
              },
              {
                "_comment": "name of the tool",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_tool_name(arguments['instance'])
              },
              {
                  "_comment": "tool description",
                  "attribute": "description",
                  "accessor": lambda arguments: _helper.get_tool_description(arguments['instance'])
              }
        ],
        [
              {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_source_agent(arguments)
              },
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.adk'
              }
        ]
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Tool input",
                "attribute": "input",
                "accessor": lambda arguments: _helper.extract_tool_input(arguments) 
            },
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from Tool",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_tool_response(arguments['result'])
            },
            {
                "attribute": "error_code",
                "accessor": lambda arguments: get_error_message(arguments)
            },
          ]
        }
      ]
}



import logging
from types import SimpleNamespace
from opentelemetry.context import attach, set_value, detach
from opentelemetry.trace import Tracer
from monocle_apptrace.instrumentation.common.constants import AGENT_PREFIX_KEY
from monocle_apptrace.instrumentation.common.span_handler import (
    SpanHandler as BaseSpanHandler,
)
from monocle_apptrace.instrumentation.common.utils import with_tracer_wrapper
from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper
from monocle_apptrace.instrumentation.metamodel.agents._helper import (
    AGENTS_AGENT_NAME_KEY,
    DELEGATION_NAME_PREFIX,
    get_runner_agent_name,
)
from monocle_apptrace.instrumentation.metamodel.agents.entities.inference import (
    AGENT_DELEGATION,
)

logger = logging.getLogger(__name__)


@with_tracer_wrapper
def constructor_wrapper(
    tracer: Tracer,
    handler: BaseSpanHandler,
    to_wrap,
    wrapped,
    instance,
    source_path,
    args,
    kwargs,
):

    original_func = kwargs.get("on_invoke_tool", None)
    result = None
    mcp_url = None
    # kwargs.get("on_invoke_tool").args[0].params["url"]
    if (
        kwargs.get("on_invoke_tool")
        and hasattr(kwargs.get("on_invoke_tool"), "args")
        and len(kwargs.get("on_invoke_tool").args) > 0
        and hasattr(kwargs.get("on_invoke_tool").args[0], "params")
    ):
        mcp_url = kwargs.get("on_invoke_tool").args[0].params.get("url", None)
        if mcp_url:
            logger.debug(f"Using MCP URL: {mcp_url}")
    tool_instance = SimpleNamespace(
        name=kwargs.get("name", "unknown_tool"),
        description=kwargs.get("description", "No description provided"),
    )
    if original_func and not getattr(original_func, "_monocle_wrapped", False):
        # Now wrap the function with our instrumentation

        async def wrapped_func(*func_args, **func_kwargs):
            token = None
            try:
                if mcp_url:
                    token = attach(set_value("mcp.url", mcp_url))
                # Use the handler to create spans when the decorated function is called
                return await atask_wrapper(
                    tracer=tracer, handler=handler, to_wrap=to_wrap
                )(
                    wrapped=original_func,
                    instance=tool_instance,
                    source_path=source_path,
                    args=func_args,
                    kwargs=func_kwargs,
                )
            finally:
                if token:
                    detach(token)

        kwargs["on_invoke_tool"] = wrapped_func
        # Preserve function metadata
        wrapped_func.__name__ = getattr(wrapped, "__name__", "unknown_tool")
        wrapped_func.__doc__ = getattr(wrapped, "__doc__", "")
        # mark function as wrapped
        setattr(wrapped_func, "_monocle_wrapped", True)

    result = wrapped(*args, **kwargs)
    return result


@with_tracer_wrapper
def handoff_constructor_wrapper(
    tracer: Tracer,
    handler: BaseSpanHandler,
    to_wrap,
    wrapped,
    instance,
    source_path,
    args,
    kwargs,
):

    original_func = kwargs.get("on_invoke_handoff", None)
    result = None
    tool_instance = SimpleNamespace(
        name=kwargs.get("name", "unknown_handoff"),
        description=kwargs.get("description", "No description provided"),
    )
    if original_func and not getattr(original_func, "_monocle_wrapped", False):
        # Now wrap the function with our instrumentation
        async def wrapped_func(*func_args, **func_kwargs):
            # Use the handler to create spans when the decorated function is called
            return await atask_wrapper(tracer=tracer, handler=handler, to_wrap=to_wrap)(
                wrapped=original_func,
                instance=tool_instance,
                source_path=source_path,
                args=func_args,
                kwargs=func_kwargs,
            )

        kwargs["on_invoke_handoff"] = wrapped_func
        # Preserve function metadata
        wrapped_func.__name__ = getattr(wrapped, "__name__", "unknown_handoff")
        wrapped_func.__doc__ = getattr(wrapped, "__doc__", "")
        # mark function as wrapped
        setattr(wrapped_func, "_monocle_wrapped", True)

    result = wrapped(*args, **kwargs)
    return result


class AgentsSpanHandler(BaseSpanHandler):
    """Span handler for OpenAI Agents SDK."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.agent_context_token = None

    def set_agent_context(self, to_wrap, wrapped, instance, args, kwargs):
        """Set the agent context for tracking across calls."""
        try:
            # For Runner.run, the agent is the first argument
            if len(args) > 0:
                agent = args[0]
                agent_name = get_runner_agent_name(agent)
                if agent_name:
                    self.agent_context_token = attach(
                        set_value(AGENTS_AGENT_NAME_KEY, agent_name)
                    )
        except Exception as e:
            logger.warning("Warning: Error setting agent context: %s", str(e))

    def clear_agent_context(self):
        """Clear the agent context."""
        if self.agent_context_token:
            detach(self.agent_context_token)
            self.agent_context_token = None

    def pre_task_processing(
        self, to_wrap, wrapped, instance, args, kwargs, *args1, **kwargs1
    ):
        """Pre-processing for agent tasks."""
        self.set_agent_context(to_wrap, wrapped, instance, args, kwargs)
        context = set_value(AGENT_PREFIX_KEY, DELEGATION_NAME_PREFIX)
        attach(context)
        return super().pre_task_processing(
            to_wrap, wrapped, instance, args, kwargs, *args1, **kwargs1
        )

    def post_task_processing(
        self, to_wrap, wrapped, instance, args, kwargs, result, ex, span, parent_span
    ):
        """Post-processing for agent tasks."""
        self.clear_agent_context()
        context = set_value(AGENT_PREFIX_KEY, None)
        attach(context)
        return super().post_task_processing(
            to_wrap, wrapped, instance, args, kwargs, result, ex, span, parent_span
        )




from monocle_apptrace.instrumentation.common.constants import SPAN_TYPES
from monocle_apptrace.instrumentation.common.utils import get_error_message, resolve_from_alias
from monocle_apptrace.instrumentation.metamodel.hugging_face import _helper


INFERENCE = {
    "type": SPAN_TYPES.INFERENCE,
    "attributes": [
        [
            {
                "_comment": "provider type, name, deployment, inference_endpoint",
                "attribute": "type",
                "accessor": lambda arguments: "inference.huggingface"
            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: "huggingface"
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: "https://api-inference.huggingface.co/v1/"
            }
        ],
        [
            {
                "_comment": "LLM Model (repo ID on Hugging Face hub)",
                "attribute": "name",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["kwargs"],
                    ["model", "model_name", "endpoint_name", "deployment_name"]
                )
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: (
                    "model.llm." +
                    resolve_from_alias(
                        arguments["kwargs"],
                        ["model", "model_name", "endpoint_name", "deployment_name"]
                    )
                )
            }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is instruction and user query to LLM",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_messages(arguments["kwargs"])
                }
            ]
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "_comment": "result from Hugging Face inference",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments)
                }
            ]
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata usage from LLM, includes token counts",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(
                        arguments.get("result"),
                        include_token_counts=True  # new flag for streaming handling
                    )
                },
                {
                    "_comment": "finish reason from Hugging Face response",
                    "attribute": "finish_reason",
                    "accessor": lambda arguments: _helper.extract_finish_reason(arguments)
                },
                {
                    "_comment": "finish type mapped from finish reason",
                    "attribute": "finish_type",
                    "accessor": lambda arguments: _helper.map_finish_reason_to_finish_type(_helper.extract_finish_reason(arguments))
                },
                {
                    "attribute": "inference_sub_type",
                    "accessor": lambda arguments: _helper.agent_inference_type(arguments)
                }
            ]
        }
    ]
}


"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""
import json
import logging
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    try_option,
    get_exception_message,
    get_status_code,
)

logger = logging.getLogger(__name__)


def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if 'messages' in kwargs and len(kwargs['messages']) > 0:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})

        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []


def extract_assistant_message(arguments):
    try:
        messages = []
        status = get_status_code(arguments)
        if status == 'success' or status == 'completed':
            response = arguments["result"]
            if (response is not None and hasattr(response, "choices") and len(response.choices) > 0):
                if hasattr(response.choices[0], "message"):
                    role = (
                        response.choices[0].message.role
                        if hasattr(response.choices[0].message, "role")
                        else "assistant"
                    )
                    messages.append({role: response.choices[0].message.content})
            return get_json_dumps(messages[0]) if messages else ""
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error

    except (IndexError, AttributeError) as e:
        logger.warning(
            "Warning: Error occurred in extract_assistant_message: %s", str(e)
        )
        return None

def extract_provider_name(url):
    """Extract host from a URL string (e.g., https://api.openai.com/v1/ -> api.openai.com)"""
    if not url:
        return None
    return url.split("//")[-1].split("/")[0]

def resolve_from_alias(my_map, alias):
    """Find a alias that is not none from list of aliases"""

    for i in alias:
        if i in my_map.keys():
            return my_map[i]
    return None


def update_span_from_llm_response(response):
    meta_dict = {}
    token_usage = None
    if response is not None:
        if token_usage is None and hasattr(response, "usage") and response.usage is not None:
            token_usage = response.usage
        elif token_usage is None and hasattr(response, "response_metadata"):
            token_usage = getattr(response.response_metadata, "token_usage", None) \
                if hasattr(response.response_metadata, "token_usage") \
                else response.response_metadata.get("token_usage", None)
        if token_usage is not None:
            meta_dict.update({"completion_tokens": getattr(token_usage, "completion_tokens", None) or getattr(token_usage, "output_tokens", None)})
            meta_dict.update({"prompt_tokens": getattr(token_usage, "prompt_tokens", None) or getattr(token_usage, "input_tokens", None)})
            meta_dict.update({"total_tokens": getattr(token_usage, "total_tokens")})
    return meta_dict


from monocle_apptrace.instrumentation.common.constants import AGENT_REQUEST_SPAN_NAME, SPAN_SUBTYPES, SPAN_TYPES
from monocle_apptrace.instrumentation.metamodel.llamaindex import (
    _helper,
)

AGENT = {
    "type": SPAN_TYPES.AGENTIC_INVOCATION,
    "subtype": SPAN_SUBTYPES.ROUTING,
    "attributes": [
        [
            {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.llamaindex'
            },
            {
                "_comment": "Agent name, type and Tools.",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_agent_name(arguments['instance'])
            },
            {
                "_comment": "agent description",
                "attribute": "description",
                "accessor": lambda arguments: _helper.get_agent_description(arguments['instance'])
            }
        ]

    ],
    "events": [
        {"name": "data.input",
         "attributes": [

             {
                 "_comment": "this is instruction and user query to LLM",
                 "attribute": "input",
                 "accessor": lambda arguments: _helper.extract_agent_input(arguments['args'])
             }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is response from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_agent_response(arguments)
                }
            ]
        }
    ]
} 

AGENT_REQUEST = {
    "type": AGENT_REQUEST_SPAN_NAME,
    "subtype": SPAN_SUBTYPES.PLANNING,
    "attributes": [
        [
              {
                "_comment": "agent request type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.llamaindex'
              }
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is request to LLM",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_agent_request_input(arguments['kwargs'])
                }
            ]
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is response from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_agent_request_output(arguments)
                }
            ]
        }
    ]
}

TOOLS = {
      "type": SPAN_TYPES.AGENTIC_TOOL_INVOCATION,
      "subtype": SPAN_SUBTYPES.ROUTING,
      "attributes": [
        [
              {
                "_comment": "tool type",
                "attribute": "type",
                "accessor": lambda arguments:'tool.llamaindex'
              },
              {
                "_comment": "name of the tool",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_tool_name(arguments['args'], arguments['instance'])
              },
              {
                  "_comment": "tool description",
                  "attribute": "description",
                  "accessor": lambda arguments: _helper.get_tool_description(arguments)
              }
        ],
        [             
              {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_source_agent(arguments['parent_span'])
              },          
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.llamaindex'
              }
        ]
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Tool input",
                "attribute": "input",
                "accessor": lambda arguments: _helper.extract_tool_args(arguments) 
            }
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from Tool",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_tool_response(arguments['result'])
             }
          ]
        }
      ]
}

AGENT_DELEGATION = {
    "type": SPAN_TYPES.AGENTIC_DELEGATION,
    "subtype": SPAN_SUBTYPES.ROUTING,
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.llamaindex'
              },
              {
                "_comment": "name of the agent",
                "attribute": "from_agent",
                "accessor": lambda arguments: _helper.get_source_agent(arguments['parent_span'])
              },
              {
                "_comment": "name of the agent called",
                "attribute": "to_agent",
                "accessor": lambda arguments: _helper.get_target_agent(arguments['result'])
              }
        ]
      ]
}


from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper
from monocle_apptrace.instrumentation.metamodel.langgraph.entities.inference import (
    AGENT,
    TOOLS,
)

LANGGRAPH_METHODS = [
    {
        "package": "langgraph.graph.state",
        "object": "CompiledStateGraph",
        "method": "invoke",
        "wrapper_method": task_wrapper,
        "span_handler": "langgraph_agent_handler",
        "scope_name": "agent.invocation",
        "output_processor": AGENT,
    },
    {
        "package": "langgraph.graph.state",
        "object": "CompiledStateGraph",
        "method": "ainvoke",
        "wrapper_method": atask_wrapper,
        "span_handler": "langgraph_agent_handler",
        "scope_name": "agent.invocation",
        "output_processor": AGENT,
    },
    {
        "package": "langchain_core.tools.base",
        "object": "BaseTool",
        "method": "run",
        "wrapper_method": task_wrapper,
        "span_handler": "langgraph_tool_handler",
        "output_processor": TOOLS,
    },
    {
        "package": "langchain_core.tools.base",
        "object": "BaseTool",
        "method": "arun",
        "wrapper_method": atask_wrapper,
        "span_handler": "langgraph_tool_handler",
        "output_processor": TOOLS,
    },
]



from monocle_apptrace.instrumentation.common.constants import SPAN_TYPES
from monocle_apptrace.instrumentation.metamodel.litellm import (
    _helper,
)
from monocle_apptrace.instrumentation.common.utils import (
    get_error_message,
    resolve_from_alias,
    get_llm_type,
)
INFERENCE = {
    "type": SPAN_TYPES.INFERENCE,
    "attributes": [
        [
            {
                "_comment": "provider type ,name , deployment , inference_endpoint",
                "attribute": "type",
                "accessor": lambda arguments: "inference."
                + (get_llm_type(arguments['instance']) or 'generic')
            },
            {
                "attribute": "provider_name",
                "accessor": lambda arguments: _helper.extract_provider_name(
                    resolve_from_alias(
                        arguments["kwargs"],
                        ["azure_endpoint", "api_base", "endpoint"],
                    )
                ),
            },
            {
                "attribute": "deployment",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["kwargs"].__dict__,
                    [
                        "engine",
                        "azure_deployment",
                        "deployment_name",
                        "deployment_id",
                        "deployment",
                    ],
                ),
            },
            {
                "attribute": "inference_endpoint",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["kwargs"],
                    ["azure_endpoint", "api_base", "endpoint"],
                )
            },
        ],
        [
            {
                "_comment": "LLM Model",
                "attribute": "name",
                "accessor": lambda arguments: resolve_from_alias(
                    arguments["kwargs"],
                    ["model", "model_name", "endpoint_name", "deployment_name"],
                ),
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: "model.llm."
                + resolve_from_alias(
                    arguments["kwargs"],
                    ["model", "model_name", "endpoint_name", "deployment_name"],
                ),
            },
        ],
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is instruction and user query to LLM",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_messages(
                        arguments["kwargs"]
                    ),
                }
            ],
        },
        {
            "name": "data.output",
            "attributes": [

                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_assistant_message(arguments),
                }
            ],
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata usage from LLM",
                    "accessor": lambda arguments: _helper.update_span_from_llm_response(
                        arguments["result"]
                    ),
                }
            ],
        },
    ],
}



"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import json
import logging
from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    get_keys_as_tuple,
    get_nested_value,
    get_status_code,
    try_option,
    get_exception_message,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_mistral_finish_reason_to_finish_type
from monocle_apptrace.instrumentation.common.constants import AGENT_PREFIX_KEY, INFERENCE_AGENT_DELEGATION, INFERENCE_TURN_END, INFERENCE_TOOL_CALL

logger = logging.getLogger(__name__)


def extract_provider_name(instance):
    provider_url: Option[str] = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)

def update_input_span_events(kwargs):
    """Extract embedding input for spans"""
    if "inputs" in kwargs and isinstance(kwargs["inputs"], list):
        # Join multiple strings into one
        return " | ".join(kwargs["inputs"])
    elif "inputs" in kwargs and isinstance(kwargs["inputs"], str):
        return kwargs["inputs"]
    return ""

def update_output_span_events(results):
    """Extract embedding output for spans"""
    try:
        if hasattr(results, "data") and isinstance(results.data, list):
            embeddings = results.data
            # just return the indices, not full vectors
            embedding_summaries = [
                f"index={e.index}, dim={len(e.embedding)}"
                for e in embeddings
            ]
            output = "\n".join(embedding_summaries)
            if len(output) > 200:
                output = output[:200] + "..."
            return output
    except Exception as e:
        logger.warning("Error in update_output_span_events: %s", str(e))
    return ""

def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if inference_endpoint.is_none() and "meta" in instance.client.__dict__:
        inference_endpoint = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))


def dummy_method(arguments):
    pass


def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if "system" in kwargs and isinstance(kwargs["system"], str):
            messages.append({"system": kwargs["system"]})
        if 'messages' in kwargs and kwargs['messages']:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})
        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []


def get_exception_status_code(arguments):
    exc = arguments.get("exception")
    if exc is not None and hasattr(exc, "status_code"):
        if exc.status_code == 401:
            return "unauthorized"
        elif exc.status_code == 403:
            return "forbidden"
        elif exc.status_code == 404:
            return "not_found"
        else:
            return str(exc.status_code)
    elif exc is not None:
        return "error"
    else:
        return "success"


def extract_assistant_message(arguments):
    """
    Extract the assistant message from a Mistral response or stream chunks.
    Returns a JSON string like {"assistant": "<text>"}.
    """
    try:
        result = arguments.get("result") if isinstance(arguments, dict) else arguments
        if result is None:
            return ""

        # Handle full response
        if hasattr(result, "choices") and result.choices:
            msg_obj = result.choices[0].message
            return get_json_dumps({msg_obj.role: msg_obj.content})

        # Handle streaming: result might be a list of CompletionEvent chunks
        if isinstance(result, list):
            content = []
            for chunk in result:
                if hasattr(chunk, "data") and hasattr(chunk.data, "choices") and chunk.data.choices:
                    choice = chunk.data.choices[0]
                    if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
                        content.append(choice.delta.content or "")
            return get_json_dumps({"assistant": "".join(content)})

        return ""

    except Exception as e:
        logger.warning("Warning in extract_assistant_message: %s", str(e))
        return ""


'''def update_span_from_llm_response(response):
    meta_dict = {}
    if response is not None and hasattr(response, "usage"):
        token_usage = getattr(response, "usage", None) or getattr(response, "response_metadata", {}).get("token_usage")
        if token_usage is not None:
            meta_dict.update({"completion_tokens": getattr(response.usage, "output_tokens", 0)})
            meta_dict.update({"prompt_tokens": getattr(response.usage, "input_tokens", 0)})
            meta_dict.update({"total_tokens": getattr(response.usage, "input_tokens", 0) + getattr(response.usage, "output_tokens", 0)})
    return meta_dict'''

def update_span_from_llm_response(result, include_token_counts=False):
    tokens = {
        "completion_tokens": getattr(result, "completion_tokens", 0),
        "prompt_tokens": getattr(result, "prompt_tokens", 0),
        "total_tokens": getattr(result, "total_tokens", 0),
    } if include_token_counts else {}
    # Add other metadata fields like finish_reason, etc.
    return {**tokens, "inference_sub_type": "turn_end"}


def extract_finish_reason(arguments):
    """
    Extract stop_reason from a Mistral response or stream chunks.
    Works for both streaming (list of chunks) and full responses.
    """
    try:
        response = arguments.get("result") if isinstance(arguments, dict) else arguments
        if response is None:
            return None

        # Handle full response: single object with stop_reason
        if hasattr(response, "stop_reason") and response.stop_reason:
            return response.stop_reason

        # Handle streaming: list of chunks, last chunk may have finish_reason
        if isinstance(response, list):
            for chunk in reversed(response):
                if hasattr(chunk, "data") and hasattr(chunk.data, "choices") and chunk.data.choices:
                    fr = getattr(chunk.data.choices[0], "finish_reason", None)
                    if fr is not None:
                        return fr

    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None

    return None


def map_finish_reason_to_finish_type(finish_reason):
    """Map Mistral stop_reason to finish_type, similar to OpenAI mapping."""
    return map_mistral_finish_reason_to_finish_type(finish_reason)


def agent_inference_type(arguments):
    """Extract agent inference type from Mistral response"""
    try:
        status = get_status_code(arguments)
        if status in ('success', 'completed'):
            response = arguments.get("result")
            if response is None:
                return INFERENCE_TURN_END

            # Check if stop_reason indicates tool use
            stop_reason = getattr(response, "stop_reason", None)
            if stop_reason == "tool_use" and hasattr(response, "content") and response.content:
                agent_prefix = get_value(AGENT_PREFIX_KEY)
                for content_block in response.content:
                    if getattr(content_block, "type", None) == "tool_use" and hasattr(content_block, "name"):
                        if agent_prefix and content_block.name.startswith(agent_prefix):
                            return INFERENCE_AGENT_DELEGATION
                return INFERENCE_TOOL_CALL

            # Fallback: check the extracted message for tool content
            assistant_message = extract_assistant_message(arguments)
            if assistant_message:
                try:
                    message = json.loads(assistant_message)
                    assistant_content = message.get("assistant", "") if isinstance(message, dict) else ""
                    agent_prefix = get_value(AGENT_PREFIX_KEY)
                    if agent_prefix and agent_prefix in assistant_content:
                        return INFERENCE_AGENT_DELEGATION
                except (json.JSONDecodeError, TypeError):
                    agent_prefix = get_value(AGENT_PREFIX_KEY)
                    if agent_prefix and agent_prefix in assistant_message:
                        return INFERENCE_AGENT_DELEGATION

        return INFERENCE_TURN_END

    except Exception as e:
        logger.warning("Warning: Error occurred in agent_inference_type: %s", str(e))
        return INFERENCE_TURN_END



from monocle_apptrace.instrumentation.common.constants import SPAN_TYPES
from monocle_apptrace.instrumentation.metamodel.langchain import (
    _helper,
)

RETRIEVAL = {
    "type": SPAN_TYPES.RETRIEVAL,
    "attributes": [
        [
            {
                "_comment": "vector store name and type",
                "attribute": "name",
                "accessor": lambda arguments: type(arguments['instance'].vectorstore).__name__
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'vectorstore.' + type(arguments['instance'].vectorstore).__name__
            },
            {
                "attribute": "deployment",
                "accessor": lambda arguments: _helper.extract_vectorstore_deployment(
                    arguments['instance'].vectorstore.__dict__)
            }
        ],
        [
            {
                "_comment": "embedding model name and type",
                "attribute": "name",
                "accessor": lambda arguments: _helper.resolve_from_alias(arguments['instance'].vectorstore.embeddings.__dict__,['endpoint_name','model_id','model'])
            },
            {
                "attribute": "type",
                "accessor": lambda arguments: 'model.embedding.' + _helper.resolve_from_alias(arguments['instance'].vectorstore.embeddings.__dict__,['endpoint_name','model_id','model'])
            }
        ]
    ],
    "events": [
        {"name": "data.input",
         "attributes": [

             {
                 "_comment": "this is instruction and user query to LLM",
                 "attribute": "input",
                 "accessor": lambda arguments: _helper.update_input_span_events(arguments['args'])
             }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.update_output_span_events(arguments['result'])
                }
            ]
        }
    ]
}



"""
Helper functions for the ADK (Agent Development Kit) metamodel instrumentation.
This module provides utility functions to extract various attributes from agent and tool instances.
"""

from ast import arguments
from typing import Any, Dict, Optional
from monocle_apptrace.instrumentation.metamodel.finish_types import map_adk_finish_reason_to_finish_type

def get_model_name(args):
    return args[0].model if hasattr(args[0], 'model') else None

def get_inference_type(arguments):
    """ Find inference type from argument """
    return 'inference.gemini' ## TBD verify non-gemini inference types

def extract_inference_endpoint(instance):
    """ Get inference service end point"""
    if hasattr(instance,'api_client') and hasattr(instance.api_client, '_api_client'):
        if hasattr(instance.api_client._api_client._http_options,'base_url'):
            return instance.api_client._api_client._http_options.base_url
    return None

def extract_message(arguments):
    return str(arguments['args'][0].contents)

def extract_assistant_message(arguments):
    return str(arguments['result'].content.parts)

def update_span_from_llm_response(response, instance):
    meta_dict = {}
    if response is not None and hasattr(response, "usage_metadata") and response.usage_metadata is not None:
        token_usage = response.usage_metadata
        if token_usage is not None:
            meta_dict.update({"completion_tokens": token_usage.candidates_token_count})
            meta_dict.update({"prompt_tokens": token_usage.prompt_token_count })
            meta_dict.update({"total_tokens": token_usage.total_token_count})
    return meta_dict

def extract_finish_reason(arguments):
    if arguments["exception"] is not None:
            return None
    if hasattr(arguments['result'], 'error_code'):
        return arguments['result'].error_code
    return None

def map_finish_reason_to_finish_type(finish_reason:str):
    return map_adk_finish_reason_to_finish_type(finish_reason)

def get_agent_name(instance: Any) -> str:
    """
    Extract the name of the agent from the given instance.

    Args:
        instance: The agent instance to extract name from

    Returns:
        str: The name of the agent, or a default value if not found
    """
    return getattr(instance, 'name', 'unknown_agent')

def get_agent_description(instance: Any) -> str:
    """
    Extract the description of the agent from the given instance.

    Args:
        instance: The agent instance to extract description from

    Returns:
        str: The description of the agent, or a default value if not found
    """
    return getattr(instance, 'description', 'No description available')


def extract_agent_input(arguments: Dict[str, Any]) -> Any:
    """
    Extract the input data from agent arguments.

    Args:
        arguments: Dictionary containing agent call arguments

    Returns:
        Any: The extracted input data
    """
    return [arguments['args'][0].user_content.parts[0].text]

def extract_agent_request_input(arguments: Dict[str, Any]) -> Any:
    """
    Extract the input data from agent request.

    Args:
        arguments: Dictionary containing agent call arguments

    Returns:
        Any: The extracted input data
    """
    return [arguments['kwargs']['new_message'].parts[0].text] if 'new_message' in arguments['kwargs'] else []

def extract_agent_response(result: Any) -> Any:
    """
    Extract the response data from agent result.

    Args:
        result: The result returned by the agent

    Returns:
        Any: The extracted response data
    """
    if result:
        return str(result.content.parts[0].text)
    else:
        return ""

def get_tool_name(instance: Any) -> str:
    """
    Extract the name of the tool from the given instance.

    Args:
        instance: The tool instance to extract name from

    Returns:
        str: The name of the tool, or a default value if not found
    """
    return getattr(instance, 'name', getattr(instance, '__name__', 'unknown_tool'))

def get_tool_description(instance: Any) -> str:
    """
    Extract the description of the tool from the given instance.

    Args:
        instance: The tool instance to extract description from

    Returns:
        str: The description of the tool, or a default value if not found
    """
    return getattr(instance, 'description', getattr(instance, '__doc__', 'No description available'))

def get_source_agent(arguments) -> str:
    """
    Get the name of the source agent (the agent that is calling a tool or delegating to another agent).

    Returns:
        str: The name of the source agent
    """
    return arguments['kwargs']['tool_context'].agent_name

def get_delegating_agent(arguments) -> str:
    """
    Get the name of the delegating agent (the agent that is delegating a task to another agent).

    Args:
        arguments: Dictionary containing agent call arguments 
    Returns:
        str: The name of the delegating agent
    """
    from_agent = arguments['args'][0].agent.name if hasattr(arguments['args'][0], 'agent') else None
    if from_agent is not None:
        if get_agent_name(arguments['instance']) == from_agent:
            return None
    return from_agent

def should_skip_delegation(arguments):
    """
    Determine whether to skip the delegation based on the arguments.

    Args:
        arguments: Dictionary containing agent call arguments

    Returns:
        bool: True if delegation should be skipped, False otherwise
    """
    return get_delegating_agent(arguments) is None

def extract_tool_input(arguments: Dict[str, Any]) -> Any:
    """
    Extract the input data from tool arguments.

    Args:
        arguments: Dictionary containing tool call arguments

    Returns:
        Any: The extracted input data
    """
    return [str(arguments['kwargs'].get('args'))]

def extract_tool_response(result: Any) -> Any:
    """
    Extract the response data from tool result.

    Args:
        result: The result returned by the tool

    Returns:
        Any: The extracted response data
    """
    return str(result)

def get_target_agent(instance: Any) -> str:
    """
    Extract the name of the target agent (the agent being called/delegated to).

    Args:
        instance: The target agent instance

    Returns:
        str: The name of the target agent
    """
    return getattr(instance, 'name', getattr(instance, '__name__', 'unknown_target_agent'))



from monocle_apptrace.instrumentation.common.constants import SPAN_SUBTYPES, SPAN_TYPES
from monocle_apptrace.instrumentation.common.utils import get_error_message
from monocle_apptrace.instrumentation.metamodel.adk import _helper
from monocle_apptrace.instrumentation.common.utils import get_error_message
AGENT = {
      "type": SPAN_TYPES.AGENTIC_INVOCATION,
      "subtype": SPAN_SUBTYPES.ROUTING,
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.adk'
              },
              {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_agent_name(arguments['instance'])
              },
              {
                "_comment": "agent description",
                "attribute": "description",
                "accessor": lambda arguments: _helper.get_agent_description(arguments['instance'])
              },
              {
                "_comment": "delegating agent name",
                "attribute": "from_agent",
                "accessor": lambda arguments: _helper.get_delegating_agent(arguments)
            }
        ]
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Agent input",
                "attribute": "input",
                "accessor": lambda arguments: _helper.extract_agent_input(arguments)
            }
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
            },
            {
                "_comment": "this is response from LLM",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_agent_response(arguments['result'])
            },
            {
                "attribute": "error_code",
                "accessor": lambda arguments: get_error_message(arguments)
            }
          ]
        }
      ]
    }

REQUEST = {
      "type": "agentic.request",
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.adk'
              }
        ],
      ],
      "events": [
        {
          "name":"data.input",
          "attributes": [
            {
                "_comment": "this is Agent input",
                "attribute": "input",
                "accessor": lambda arguments: _helper.extract_agent_request_input(arguments)
            }
          ]
        },
        {
          "name":"data.output",
          "attributes": [
            {
                "_comment": "this is response from LLM",
                "attribute": "response",
                "accessor": lambda arguments: _helper.extract_agent_response(arguments['result'])
            }
          ]
        }
      ]
}

DELEGATION = {
      "type": "agentic.delegation",
      "should_skip": lambda arguments: _helper.should_skip_delegation(arguments),
      "attributes": [
        [
              {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments:'agent.adk'
              },
              {
                "_comment": "name of the agent",
                "attribute": "from_agent",
                "accessor": lambda arguments: _helper.get_delegating_agent(arguments)
              },
              {
                "_comment": "name of the agent called",
                "attribute": "to_agent",
                "accessor": lambda arguments: _helper.get_agent_name(arguments['instance'])
              }
        ]
      ]
}



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.hugging_face.entities.inference import (
    INFERENCE,
)

HUGGING_FACE_METHODS = [
    {
        "package": "huggingface_hub",
        "object": "InferenceClient",
        "method": "chat_completion",  # sync
        "wrapper_method": task_wrapper,
        "span_handler": "non_framework_handler",
        "output_processor": INFERENCE,
    },
    {
        "package": "huggingface_hub",
        "object": "AsyncInferenceClient",
        "method": "chat_completion",  # async
        "wrapper_method": atask_wrapper,
        "span_handler": "non_framework_handler",
        "output_processor": INFERENCE,
    },
]



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper
from monocle_apptrace.instrumentation.metamodel.mistral.entities.inference import MISTRAL_INFERENCE
from monocle_apptrace.instrumentation.metamodel.mistral.entities.retrieval import MISTRAL_RETRIEVAL

MISTRAL_METHODS = [
    {
        "package": "mistralai.chat",          # where Chat is defined
        "object": "Chat",                     # class name
        "method": "complete",                 # the sync method
        "span_handler": "non_framework_handler",
        "wrapper_method": task_wrapper,
        "output_processor": MISTRAL_INFERENCE
    },
    {
        "package": "mistralai.chat",          # where Chat is defined
        "object": "Chat",                     # class name
        "method": "complete_async",           # the async method
        "span_handler": "non_framework_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": MISTRAL_INFERENCE
    },
    {
        "package": "mistralai.chat",
        "object": "Chat",
        "method": "stream",              # sync streaming
        "span_handler": "non_framework_handler",
        "wrapper_method": task_wrapper,
        "output_processor": MISTRAL_INFERENCE,
    },
    {
        "package": "mistralai.chat",
        "object": "Chat",
        "method": "stream_async",        # async streaming
        "span_handler": "non_framework_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": MISTRAL_INFERENCE,
    },
    {
        "package": "mistralai.embeddings",    # where Embeddings is defined
        "object": "Embeddings",               # sync embeddings client
        "method": "create",                   # sync create
        "span_handler": "non_framework_handler",
        "wrapper_method": task_wrapper,
        "output_processor": MISTRAL_RETRIEVAL
    },
    {
        "package": "mistralai.embeddings",    # where Embeddings is defined
        "object": "AsyncEmbeddings",          # async embeddings client
        "method": "create",                   # async create
        "span_handler": "non_framework_handler",
        "wrapper_method": atask_wrapper,
        "output_processor": MISTRAL_RETRIEVAL
    }
]

 





from monocle_apptrace.instrumentation.common.constants import SPAN_TYPES
from monocle_apptrace.instrumentation.metamodel.flask import _helper
FLASK_HTTP_PROCESSOR = {
    "type": SPAN_TYPES.HTTP_PROCESS,
    "attributes": [
        [
            {
                "_comment": "request method, request URI",
                "attribute": "method",
                "accessor": lambda arguments: _helper.get_method(arguments['args'])
            },
            {
                "_comment": "request method, request URI",
                "attribute": "route",
                "accessor": lambda arguments: _helper.get_route(arguments['args'])
            },
            {
                "_comment": "request method, request URI",
                "attribute": "url",
                "accessor": lambda arguments: _helper.get_url(arguments['args'])
            },
        ]
    ]
}

FLASK_RESPONSE_PROCESSOR = {
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "route params",
                    "attribute": "params",
                    "accessor": lambda arguments: _helper.get_params(arguments['args'])
                }
         ]
         },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "status from HTTP response",
                    "attribute": "error_code",
                    "accessor": lambda arguments: _helper.extract_status(arguments)
                },
                {
                    "_comment": "this is result from LLM",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_response(arguments['instance'])
                }
            ]
        }
    ]
}


# Azure environment constants
AZURE_ML_ENDPOINT_ENV_NAME = "AZUREML_ENTRY_SCRIPT"
AZURE_FUNCTION_WORKER_ENV_NAME = "FUNCTIONS_WORKER_RUNTIME"
AZURE_APP_SERVICE_ENV_NAME = "WEBSITE_SITE_NAME"
AWS_LAMBDA_ENV_NAME = "AWS_LAMBDA_RUNTIME_API"
GITHUB_CODESPACE_ENV_NAME = "CODESPACES"

AWS_LAMBDA_FUNCTION_IDENTIFIER_ENV_NAME = "AWS_LAMBDA_FUNCTION_NAME"
AZURE_FUNCTION_IDENTIFIER_ENV_NAME = "WEBSITE_SITE_NAME"
AZURE_APP_SERVICE_IDENTIFIER_ENV_NAME = "WEBSITE_DEPLOYMENT_ID"
GITHUB_CODESPACE_IDENTIFIER_ENV_NAME = "GITHUB_REPOSITORY"


# Azure naming reference can be found here
# https://learn.microsoft.com/en-us/azure/cloud-adoption-framework/ready/azure-best-practices/resource-abbreviations
# https://docs.aws.amazon.com/resource-explorer/latest/userguide/supported-resource-types.html#services-lookoutmetrics
AZURE_FUNCTION_NAME = "azure_func"
AZURE_APP_SERVICE_NAME = "azure_webapp"
AZURE_ML_SERVICE_NAME = "azure_ml"
AWS_LAMBDA_SERVICE_NAME = "aws_lambda"
GITHUB_CODESPACE_SERVICE_NAME = "github_codespace"

# Env variables to identify infra service type
service_type_map = {
    AZURE_ML_ENDPOINT_ENV_NAME: AZURE_ML_SERVICE_NAME,
    AZURE_APP_SERVICE_ENV_NAME: AZURE_APP_SERVICE_NAME,
    AZURE_FUNCTION_WORKER_ENV_NAME: AZURE_FUNCTION_NAME,
    AWS_LAMBDA_ENV_NAME: AWS_LAMBDA_SERVICE_NAME,
    GITHUB_CODESPACE_ENV_NAME: GITHUB_CODESPACE_SERVICE_NAME,
}

# Env variables to identify infra service name
service_name_map = {
    AZURE_APP_SERVICE_NAME: AZURE_APP_SERVICE_IDENTIFIER_ENV_NAME,
    AZURE_FUNCTION_NAME: AZURE_FUNCTION_IDENTIFIER_ENV_NAME,
    AZURE_ML_SERVICE_NAME: AZURE_ML_ENDPOINT_ENV_NAME,
    AWS_LAMBDA_SERVICE_NAME: AWS_LAMBDA_FUNCTION_IDENTIFIER_ENV_NAME,
    GITHUB_CODESPACE_SERVICE_NAME: GITHUB_CODESPACE_IDENTIFIER_ENV_NAME,
}


llm_type_map = {
    "sagemakerendpoint": "aws_sagemaker",
    "azureopenai": "azure_openai",
    "openai": "openai",
    "chatopenai": "openai",
    "azurechatopenai": "azure_openai",
    "bedrock": "aws_bedrock",
    "sagemakerllm": "aws_sagemaker",
    "chatbedrock": "aws_bedrock",
    "openaigenerator": "openai",
    "bedrockruntime": "aws_bedrock",
    "sagemakerruntime": "aws_sagemaker",
    "anthropic": "anthropic",
    "chatanthropic": "anthropic",
    "anthropicchatgenerator": "anthropic",
    "chatcompletionsclient": "azure_ai_inference",
    "embeddingsclient": "azure_ai_inference",
    "imageembeddingsclient": "azure_ai_inference",
    "chatbedrockconverse": "aws_bedrock",
    "googleaigeminichatgenerator": "gemini",
    "gemini": "gemini",
    "chatgooglegenerativeai": "gemini",
    "azurechatcompletion": "azure_openai",
    "openaichatcompletion": "openai",
}

MONOCLE_INSTRUMENTOR = "monocle_apptrace"
WORKFLOW_TYPE_KEY = "workflow_type"
DATA_INPUT_KEY = "data.input"
DATA_OUTPUT_KEY = "data.output"
PROMPT_INPUT_KEY = "data.input"
PROMPT_OUTPUT_KEY = "data.output"
QUERY = "input"
RESPONSE = "response"
SESSION_PROPERTIES_KEY = "session"
INFRA_SERVICE_KEY = "infra_service_name"
META_DATA = "metadata"
MONOCLE_SCOPE_NAME_PREFIX = "monocle.scope."
SCOPE_METHOD_LIST = "MONOCLE_SCOPE_METHODS"
SCOPE_METHOD_FILE = "monocle_scopes.json"
SCOPE_CONFIG_PATH = "MONOCLE_SCOPE_CONFIG_PATH"
TRACE_PROPOGATION_URLS = "MONOCLE_TRACE_PROPAGATATION_URLS"
WORKFLOW_TYPE_KEY = "monocle.workflow_type"
ADD_NEW_WORKFLOW = "monocle.add_new_workflow"
WORKFLOW_TYPE_GENERIC = "workflow.generic"
MONOCLE_SDK_VERSION = "monocle_apptrace.version"
MONOCLE_SDK_LANGUAGE = "monocle_apptrace.language"
MONOCLE_DETECTED_SPAN_ERROR = "monocle_apptrace.detected_span_error"
HTTP_SUCCESS_CODES = ("200", "201", "202", "204", "205", "206")
CHILD_ERROR_CODE = "child.error.code"

AGENT_PREFIX_KEY = "monocle.agent.prefix"

# agentic sub types
INFERENCE_AGENT_DELEGATION = "delegation"
INFERENCE_TOOL_CALL = "tool_call"
INFERENCE_TURN_END = "turn_end"

SCOPE_NAME = "scope_name"
AGENT_INVOCATION_SPAN_NAME = "agentic.invocation"
AGENT_REQUEST_SPAN_NAME = "agentic.request"

AGENTIC_SPANS = [AGENT_INVOCATION_SPAN_NAME, AGENT_REQUEST_SPAN_NAME]

# Span sub types

## OPTIONAL right next to span.type,  span.subtype:
## subtype is one perspective , are non overlapping, limitations: only one classification scheme for subtypes
# 1 planning
SPAN_SUBTYPE_PLANNING = "planning"

# 2 routing and selection INFERENCE_TOOL_CALL, INFERENCE_AGENT_DELEGATION
SPAN_SUBTYPE_ROUTING = "routing"

# 3 content processing
SPAN_SUBTYPE_CONTENT_PROCESSING = "content_processing"

# 4 content generation
SPAN_SUBTYPE_CONTENT_GENERATION = "content_generation"

# 5 communication INFERENCE_TURN_END
SPAN_SUBTYPE_COMMUNICATION = "communication"

# 6 transformations , if structured output
SPAN_SUBTYPE_TRANSFORMATIONS = "transformations"

# 7 domain specific,
SPAN_SUBTYPE_DOMAIN_SPECIFIC = "domain_specific"

# 8 generic (we may skip this property)
SPAN_SUBTYPE_GENERIC = "generic"

class SPAN_TYPES:
    GENERIC = "generic"
    AGENTIC_DELEGATION = "agentic.delegation"
    AGENTIC_TOOL_INVOCATION = "agentic.tool.invocation"
    AGENTIC_INVOCATION = "agentic.invocation"
    AGENTIC_MCP_INVOCATION = "agentic.mcp.invocation"
    AGENTIC_REQUEST = "agentic.request"

    # http.process
    HTTP_PROCESS = "http.process"
    HTTP_SEND = "http.send"

    RETRIEVAL = "retrieval"
    INFERENCE = "inference"
    INFERENCE_FRAMEWORK = "inference.framework"


class SPAN_SUBTYPES:
    PLANNING = SPAN_SUBTYPE_PLANNING
    ROUTING = SPAN_SUBTYPE_ROUTING
    CONTENT_PROCESSING = SPAN_SUBTYPE_CONTENT_PROCESSING
    CONTENT_GENERATION = SPAN_SUBTYPE_CONTENT_GENERATION
    COMMUNICATION = SPAN_SUBTYPE_COMMUNICATION
    TRANSFORMATIONS = SPAN_SUBTYPE_TRANSFORMATIONS
    DOMAIN_SPECIFIC = SPAN_SUBTYPE_DOMAIN_SPECIFIC
    GENERIC = SPAN_SUBTYPE_GENERIC


MAP_ATTRIBUTES_TO_SPAN_SUBTYPE = {
    # inference attributes
    INFERENCE_AGENT_DELEGATION: SPAN_SUBTYPES.ROUTING,
    INFERENCE_TOOL_CALL: SPAN_SUBTYPES.ROUTING,
    INFERENCE_TURN_END: SPAN_SUBTYPES.COMMUNICATION,

    # agentic span.types
    SPAN_TYPES.AGENTIC_DELEGATION: SPAN_SUBTYPES.ROUTING,
    SPAN_TYPES.AGENTIC_TOOL_INVOCATION: SPAN_SUBTYPES.ROUTING,
    SPAN_TYPES.AGENTIC_INVOCATION: SPAN_SUBTYPES.ROUTING,
    SPAN_TYPES.AGENTIC_MCP_INVOCATION: SPAN_SUBTYPES.ROUTING,
    
    # MAYBE?
    # agentic span.types
    SPAN_TYPES.AGENTIC_REQUEST: SPAN_SUBTYPES.PLANNING,
    
    
    
}



"""
This module provides utility functions for extracting system, user,
and assistant messages from various input formats.
"""

import json
import logging
from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    get_keys_as_tuple,
    get_nested_value,
    get_status_code,
    try_option,
    get_exception_message,
)
from monocle_apptrace.instrumentation.metamodel.finish_types import map_anthropic_finish_reason_to_finish_type
from monocle_apptrace.instrumentation.common.constants import AGENT_PREFIX_KEY, INFERENCE_AGENT_DELEGATION, INFERENCE_TURN_END, INFERENCE_TOOL_CALL


logger = logging.getLogger(__name__)

def extract_provider_name(instance):
    provider_url: Option[str] = try_option(getattr, instance._client.base_url, 'host')
    return provider_url.unwrap_or(None)

def extract_inference_endpoint(instance):
    inference_endpoint: Option[str] = try_option(getattr, instance._client, 'base_url').map(str)
    if inference_endpoint.is_none() and "meta" in instance.client.__dict__:
        inference_endpoint = try_option(getattr, instance.client.meta, 'endpoint_url').map(str)

    return inference_endpoint.unwrap_or(extract_provider_name(instance))

def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if "system" in kwargs and isinstance(kwargs["system"], str):
            messages.append({"system": kwargs["system"]})
        if 'messages' in kwargs and len(kwargs['messages']) >0:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})
        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def get_exception_status_code(arguments):
    if arguments['exception'] is not None and hasattr(arguments['exception'], 'status_code') and arguments['exception'].status_code is not None:
        return arguments['exception'].status_code
    elif arguments['exception'] is not None:
        return 'error'
    else:
        return 'success'

def extract_assistant_message(arguments):
    try:
        status = get_status_code(arguments)
        response = arguments["result"]
        if status == 'success':
            messages = []
            role = response.role if hasattr(response, 'role') else "assistant"
            
            # Handle tool use content blocks
            if hasattr(response, "content") and response.content:
                tools = []
                text_content = []
                
                for content_block in response.content:
                    if hasattr(content_block, "type"):
                        if content_block.type == "tool_use":
                            # Extract tool use information
                            tool_info = {
                                "tool_id": getattr(content_block, "id", ""),
                                "tool_name": getattr(content_block, "name", ""),
                                "tool_arguments": getattr(content_block, "input", "")
                            }
                            tools.append(tool_info)
                        elif content_block.type == "text":
                            # Extract text content
                            if hasattr(content_block, "text"):
                                text_content.append(content_block.text)
                
                # If we have tools, add them to the message
                if tools:
                    messages.append({"tools": tools})
                
                # If we have text content, add it to the message
                if text_content:
                    messages.append({role: " ".join(text_content)})
                
                # Fallback to original logic if no content blocks were processed
                if not messages and len(response.content) > 0:
                    if hasattr(response.content[0], "text"):
                        messages.append({role: response.content[0].text})
            
            # Return first message if list is not empty
            return get_json_dumps(messages[0]) if messages else ""
        else:
            if arguments["exception"] is not None:
                return get_exception_message(arguments)
            elif hasattr(arguments["result"], "error"):
                return arguments["result"].error

    except (IndexError, AttributeError) as e:
        logger.warning("Warning: Error occurred in extract_assistant_message: %s", str(e))
        return None

def update_span_from_llm_response(response):
    meta_dict = {}
    if response is not None and hasattr(response, "usage"):
        if hasattr(response, "usage") and response.usage is not None:
            token_usage = response.usage
        else:
            response_metadata = response.response_metadata
            token_usage = response_metadata.get("token_usage")
        if token_usage is not None:
            meta_dict.update({"completion_tokens": getattr(response.usage, "output_tokens", 0)})
            meta_dict.update({"prompt_tokens": getattr(response.usage, "input_tokens", 0)})
            meta_dict.update({"total_tokens": getattr(response.usage, "input_tokens", 0)+getattr(response.usage, "output_tokens", 0)})
    return meta_dict

def extract_finish_reason(arguments):
    """Extract stop_reason from Anthropic response (Claude)."""
    try:
        # Arguments may be a dict with 'result' or just the response object
        response = arguments.get("result") if isinstance(arguments, dict) else arguments
        if response is not None and hasattr(response, "stop_reason"):
            return response.stop_reason
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_finish_reason: %s", str(e))
        return None
    return None

def map_finish_reason_to_finish_type(finish_reason):
    """Map Anthropic stop_reason to finish_type, similar to OpenAI mapping."""
    return map_anthropic_finish_reason_to_finish_type(finish_reason)

def agent_inference_type(arguments):
    """Extract agent inference type from Anthropic response"""
    try:
        status = get_status_code(arguments)
        if status == 'success' or status == 'completed':
            response = arguments["result"]
            
            # Check if stop_reason indicates tool use
            if hasattr(response, "stop_reason") and response.stop_reason == "tool_use":
                # Check if this is agent delegation by looking at tool names
                if hasattr(response, "content") and response.content:
                    agent_prefix = get_value(AGENT_PREFIX_KEY)
                    for content_block in response.content:
                        if (hasattr(content_block, "type") and 
                            content_block.type == "tool_use" and
                            hasattr(content_block, "name")):
                            tool_name = content_block.name
                            if agent_prefix and tool_name.startswith(agent_prefix):
                                return INFERENCE_AGENT_DELEGATION
                    # If we found tool use but no agent delegation, it's a regular tool call
                    return INFERENCE_TOOL_CALL
            
            # Fallback: check the extracted message for tool content
            assistant_message = extract_assistant_message(arguments)
            if assistant_message:
                try:
                    message = json.loads(assistant_message)
                    if message and isinstance(message, dict):
                        assistant_content = message.get("assistant", "")
                        if assistant_content:
                            agent_prefix = get_value(AGENT_PREFIX_KEY)
                            if agent_prefix and agent_prefix in assistant_content:
                                return INFERENCE_AGENT_DELEGATION
                except (json.JSONDecodeError, TypeError):
                    # If JSON parsing fails, fall back to string analysis
                    agent_prefix = get_value(AGENT_PREFIX_KEY)
                    if agent_prefix and agent_prefix in assistant_message:
                        return INFERENCE_AGENT_DELEGATION
        
        return INFERENCE_TURN_END
    except Exception as e:
        logger.warning("Warning: Error occurred in agent_inference_type: %s", str(e))
        return INFERENCE_TURN_END


# pylint: disable=protected-access
from contextlib import contextmanager
import os
from typing import AsyncGenerator, Iterator, Optional
import logging
from opentelemetry.trace import Tracer
from opentelemetry.trace.propagation import _SPAN_KEY, set_span_in_context, get_current_span
from opentelemetry.trace import propagation
from opentelemetry.context import set_value, attach, detach, get_value
from opentelemetry.context import create_key, get_value, set_value
from opentelemetry.context.context import Context
from opentelemetry.trace.span import INVALID_SPAN, Span
from opentelemetry.trace.status import StatusCode

from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.utils import (
    set_scopes,
    with_tracer_wrapper,
    set_scope,
    remove_scope,
    get_current_monocle_span,
    set_monocle_span_in_context
)
from monocle_apptrace.instrumentation.common.constants import WORKFLOW_TYPE_KEY, ADD_NEW_WORKFLOW, AGENTIC_SPANS
from monocle_apptrace.instrumentation.common.scope_wrapper import monocle_trace_scope

logger = logging.getLogger(__name__)
ISOLATE_MONOCLE_SPANS = os.getenv("MONOCLE_ISOLATE_SPANS", "true").lower() == "true"

def get_auto_close_span(to_wrap, kwargs):
    try:
        if to_wrap.get("output_processor") and to_wrap.get("output_processor").get("is_auto_close"):
            return to_wrap.get("output_processor").get("is_auto_close")(kwargs)
        return True
    except Exception as e:
        logger.warning("Warning: Error occurred in get_auto_close_span: %s", str(e))
        return True

def pre_process_span(name, tracer, handler, add_workflow_span, to_wrap, wrapped, instance, args, kwargs, span, source_path):
    SpanHandler.set_default_monocle_attributes(span, source_path)
    if SpanHandler.is_root_span(span) or add_workflow_span:
        # This is a direct API call of a non-framework type
        SpanHandler.set_workflow_properties(span, to_wrap)
    else:
        SpanHandler.set_non_workflow_properties(span)
        try:
            handler.pre_task_processing(to_wrap, wrapped, instance, args, kwargs, span)
        except Exception as e:
            logger.info(f"Warning: Error occurred in pre_task_processing: {e}")

def post_process_span(handler, to_wrap, wrapped, instance, args, kwargs, return_value, span, parent_span, ex):
    if not (SpanHandler.is_root_span(span) or get_value(ADD_NEW_WORKFLOW) == True):
        try:
            if parent_span == INVALID_SPAN:
                parent_span = None
            handler.hydrate_span(to_wrap, wrapped, instance, args, kwargs, return_value, span, parent_span, ex)
        except Exception as e:
            logger.info(f"Warning: Error occurred in hydrate_span: {e}")
        
        try:
            handler.post_task_processing(to_wrap, wrapped, instance, args, kwargs, return_value, ex, span, parent_span)
        except Exception as e:
            logger.info(f"Warning: Error occurred in post_task_processing: {e}")

def get_span_name(to_wrap, instance):
    if to_wrap.get("span_name"):
        name = to_wrap.get("span_name")
    else:
        name = to_wrap.get("package", "") + "." + to_wrap.get("object", "") + "." + to_wrap.get("method", "")
    return name

def monocle_wrapper_span_processor(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, add_workflow_span, args, kwargs):
    # Main span processing logic
    name = get_span_name(to_wrap, instance)
    return_value = None
    span_status = None
    auto_close_span = get_auto_close_span(to_wrap, kwargs)
    parent_span = get_current_monocle_span()
    with start_as_monocle_span(tracer, name, auto_close_span) as span:
        pre_process_span(name, tracer, handler, add_workflow_span, to_wrap, wrapped, instance, args, kwargs, span, source_path)
        
        if SpanHandler.is_root_span(span) or add_workflow_span:
            # Recursive call for the actual span
            return_value, span_status = monocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs)
            span.set_status(StatusCode.OK)
            if not auto_close_span:
                span.end()
        else:
            ex:Exception = None
            to_wrap = get_wrapper_with_next_processor(to_wrap, handler, instance, args, kwargs)
            if has_more_processors(to_wrap):
                try:
                    return_value, span_status = monocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs)
                except Exception as e:
                    ex = e
                    raise
                finally:
                    post_process_span_internal(return_value)
            else:
                try:
                    with SpanHandler.workflow_type(to_wrap, span):
                        return_value = wrapped(*args, **kwargs)
                except Exception as e:
                    ex = e
                    raise
                finally:
                    def post_process_span_internal(ret_val):
                        post_process_span(handler, to_wrap, wrapped, instance, args, kwargs, ret_val, span, parent_span ,ex)
                        if not auto_close_span:
                            span.end()
                    if ex is None and not auto_close_span and to_wrap.get("output_processor") and to_wrap.get("output_processor").get("response_processor"):
                        to_wrap.get("output_processor").get("response_processor")(to_wrap, return_value, post_process_span_internal)
                    else:
                        post_process_span_internal(return_value)
            span_status = span.status
    return return_value, span_status

def monocle_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    return_value = None
    pre_trace_token = None
    token = None
    try:
        try:
            pre_trace_token = handler.pre_tracing(to_wrap, wrapped, instance, args, kwargs)
        except Exception as e:
            logger.info(f"Warning: Error occurred in pre_tracing: {e}")
        if to_wrap.get('skip_span', False) or handler.skip_span(to_wrap, wrapped, instance, args, kwargs):
            return_value = wrapped(*args, **kwargs)
        else:
            add_workflow_span = get_value(ADD_NEW_WORKFLOW) == True
            token = attach(set_value(ADD_NEW_WORKFLOW, False))
            try:
                with monocle_trace_scope(get_builtin_scope_names(to_wrap)):
                    return_value, span_status = monocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, add_workflow_span, args, kwargs)
            finally:
                detach(token)
        return return_value
    finally:
        try:
            handler.post_tracing(to_wrap, wrapped, instance, args, kwargs, return_value, token=pre_trace_token)
        except Exception as e:
            logger.info(f"Warning: Error occurred in post_tracing: {e}")

async def amonocle_wrapper_span_processor(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, add_workflow_span,
                                        args, kwargs):
    # Main span processing logic
    name = get_span_name(to_wrap, instance)
    return_value = None
    span_status = None
    auto_close_span = get_auto_close_span(to_wrap, kwargs)
    parent_span = get_current_monocle_span()
    with start_as_monocle_span(tracer, name, auto_close_span) as span:
        pre_process_span(name, tracer, handler, add_workflow_span, to_wrap, wrapped, instance, args, kwargs, span, source_path)
        
        if SpanHandler.is_root_span(span) or add_workflow_span:
            # Recursive call for the actual span
            return_value, span_status = await amonocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs)
            span.set_status(StatusCode.OK)
            if not auto_close_span:
                span.end()
        else:
            ex:Exception = None
            to_wrap = get_wrapper_with_next_processor(to_wrap, handler, instance, args, kwargs)
            if has_more_processors(to_wrap):
                try:
                    return_value, span_status = await amonocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs)
                except Exception as e:
                    ex = e
                    raise
                finally:
                    post_process_span_internal(return_value)
            else:
                try:
                    with SpanHandler.workflow_type(to_wrap, span):
                        return_value = await wrapped(*args, **kwargs)
                except Exception as e:
                    ex = e
                    raise
                finally:
                    def post_process_span_internal(ret_val):
                        post_process_span(handler, to_wrap, wrapped, instance, args, kwargs, ret_val, span, parent_span, ex)
                        if not auto_close_span:
                            span.end()
                    if ex is None and not auto_close_span and to_wrap.get("output_processor") and to_wrap.get("output_processor").get("response_processor"):
                        to_wrap.get("output_processor").get("response_processor")(to_wrap, return_value, post_process_span_internal)
                    else:
                        post_process_span_internal(return_value)
        span_status = span.status
    return return_value, span_status

async def amonocle_iter_wrapper_span_processor(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, add_workflow_span,
                                        args, kwargs) -> AsyncGenerator[any, None]:
    # Main span processing logic
    name = get_span_name(to_wrap, instance)
    auto_close_span = get_auto_close_span(to_wrap, kwargs)
    parent_span = get_current_monocle_span()
    last_item = None

    with start_as_monocle_span(tracer, name, auto_close_span) as span:
        pre_process_span(name, tracer, handler, add_workflow_span, to_wrap, wrapped, instance, args, kwargs, span, source_path)

        if SpanHandler.is_root_span(span) or add_workflow_span:
            # Recursive call for the actual span
            async for item in amonocle_iter_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs):
                yield item
            span.set_status(StatusCode.OK)
            if not auto_close_span:
                span.end()
        else:
            ex:Exception = None
            to_wrap = get_wrapper_with_next_processor(to_wrap, handler, instance, args, kwargs)
            if has_more_processors(to_wrap):
                try:
                    async for item in amonocle_iter_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, False, args, kwargs):
                        last_item = item
                        yield item
                except Exception as e:
                    ex = e
                    raise
                finally:
                    post_process_span(handler, to_wrap, wrapped, instance, args, kwargs, last_item, span, parent_span, ex)
            else:
                try:
                    with SpanHandler.workflow_type(to_wrap, span):
                        async for item in wrapped(*args, **kwargs):
                            last_item = item
                            yield item
                except Exception as e:
                    ex = e
                    raise
                finally:
                    def post_process_span_internal(ret_val):
                        post_process_span(handler, to_wrap, wrapped, instance, args, kwargs, ret_val, span, parent_span, ex)
                        if not auto_close_span:
                            span.end()
                    if ex is None and not auto_close_span and to_wrap.get("output_processor") and to_wrap.get("output_processor").get("response_processor"):
                        to_wrap.get("output_processor").get("response_processor")(to_wrap, None, post_process_span_internal)
                    else:
                        post_process_span_internal(last_item)
    return

async def amonocle_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    return_value = None
    token = None
    pre_trace_token = None
    try:
        try:
            pre_trace_token = handler.pre_tracing(to_wrap, wrapped, instance, args, kwargs)
        except Exception as e:
            logger.info(f"Warning: Error occurred in pre_tracing: {e}")
        if to_wrap.get('skip_span', False) or handler.skip_span(to_wrap, wrapped, instance, args, kwargs):
            return_value = await wrapped(*args, **kwargs)
        else:
            add_workflow_span = get_value(ADD_NEW_WORKFLOW) == True
            token = attach(set_value(ADD_NEW_WORKFLOW, False))
            try:
                with monocle_trace_scope(get_builtin_scope_names(to_wrap)):
                    return_value, span_status = await amonocle_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, 
                                                                        add_workflow_span, args, kwargs)
            finally:
                detach(token)
        return return_value
    finally:
        try:
            handler.post_tracing(to_wrap, wrapped, instance, args, kwargs, return_value, pre_trace_token)
        except Exception as e:
            logger.info(f"Warning: Error occurred in post_tracing: {e}")

async def amonocle_iter_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs) -> AsyncGenerator[any, None]:
    token = None
    pre_trace_token = None
    try:
        try:
            pre_trace_token = handler.pre_tracing(to_wrap, wrapped, instance, args, kwargs)
        except Exception as e:
            logger.info(f"Warning: Error occurred in pre_tracing: {e}")
        if to_wrap.get('skip_span', False) or handler.skip_span(to_wrap, wrapped, instance, args, kwargs):
            async for item in wrapped(*args, **kwargs):
                yield item
        else:
            add_workflow_span = get_value(ADD_NEW_WORKFLOW) == True
            token = attach(set_value(ADD_NEW_WORKFLOW, False))
            try:
                with monocle_trace_scope(get_builtin_scope_names(to_wrap)):
                    async for item in amonocle_iter_wrapper_span_processor(tracer, handler, to_wrap, wrapped, instance, source_path, add_workflow_span, args, kwargs):
                        yield item
            finally:
                detach(token)
        return
    finally:
        try:
            handler.post_tracing(to_wrap, wrapped, instance, args, kwargs, None, pre_trace_token)
        except Exception as e:
            logger.info(f"Warning: Error occurred in post_tracing: {e}")

@with_tracer_wrapper
def task_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    return monocle_wrapper(tracer, handler, to_wrap, wrapped, instance, source_path, args, kwargs)

@with_tracer_wrapper
async def atask_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    return await amonocle_wrapper(tracer, handler, to_wrap, wrapped, instance, source_path, args, kwargs)

@with_tracer_wrapper
async def atask_iter_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs) -> AsyncGenerator[any, None]:
    async for item in amonocle_iter_wrapper(tracer, handler, to_wrap, wrapped, instance, source_path, args, kwargs):
        yield item
    return

@with_tracer_wrapper
def scope_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    scope_name = to_wrap.get('scope_name', None)
    if scope_name:
        token = set_scope(scope_name)
    return_value = wrapped(*args, **kwargs)
    if scope_name:
        remove_scope(token)
    return return_value

@with_tracer_wrapper
async def ascope_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    scope_name = to_wrap.get('scope_name', None)
    scope_value = to_wrap.get('scope_value', None)
    token = None
    try:
        if scope_name:
            token = set_scope(scope_name, scope_value)
        return_value = await wrapped(*args, **kwargs)
        return return_value
    finally:
        if token:
            remove_scope(token)
            
@with_tracer_wrapper
def scopes_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    scope_values = to_wrap.get('scope_values', None)
    scope_values = evaluate_scope_values(args, kwargs, to_wrap, scope_values)
    token = None
    try:
        if scope_values:
            token = set_scopes(scope_values)
        return_value = wrapped(*args, **kwargs)
        return return_value
    finally:
        if token:
            remove_scope(token)

@with_tracer_wrapper
async def ascopes_wrapper(tracer: Tracer, handler: SpanHandler, to_wrap, wrapped, instance, source_path, args, kwargs):
    scope_values = to_wrap.get('scope_values', None)
    scope_values = evaluate_scope_values(args, kwargs, to_wrap, scope_values)
    token = None
    try:
        if scope_values:
            token = set_scopes(scope_values)
        return_value = await wrapped(*args, **kwargs)
        return return_value
    finally:
        if token:
            remove_scope(token)

def evaluate_scope_values(args, kwargs, to_wrap, scope_values):
    if callable(scope_values):
        try:
            scope_values = scope_values(args, kwargs)
        except Exception as e:
            logger.warning("Warning: Error occurred in evaluate_scope_values: %s", str(e))
            scope_values = None
    if isinstance(scope_values, dict):
        return scope_values
    return None

@contextmanager
def start_as_monocle_span(tracer: Tracer, name: str, auto_close_span: bool) -> Iterator["Span"]:
    """ Wrapper to OTEL start_as_current_span to isolate monocle and non monocle spans.
        This essentiall links monocle and non-monocle spans separately which is default behavior.
        It can be optionally overridden by setting the environment variable MONOCLE_ISOLATE_SPANS to false.
    """
    if not ISOLATE_MONOCLE_SPANS:
        # If not isolating, use the default start_as_current_span
        yield tracer.start_as_current_span(name, end_on_exit=auto_close_span)
        return
    original_span = get_current_span()
    monocle_span_token = attach(set_span_in_context(get_current_monocle_span()))
    with tracer.start_as_current_span(name, end_on_exit=auto_close_span) as span:
        new_monocle_token = attach(set_monocle_span_in_context(span))
        original_span_token = attach(set_span_in_context(original_span))
        yield span
        detach(original_span_token)
        detach(new_monocle_token)
    detach(monocle_span_token)

def get_builtin_scope_names(to_wrap) -> str:
    output_processor = to_wrap.get("output_processor", None)
    span_type = output_processor.get("type", None) if output_processor else None
    if span_type and span_type in AGENTIC_SPANS:
        return span_type
    return None

def get_wrapper_with_next_processor(to_wrap, handler, instance, args, kwargs):
    if has_more_processors(to_wrap):
        next_output_processor_list = to_wrap.get('output_processor_list',[]).copy()
        while len(next_output_processor_list) > 0:
            next_output_processor = next_output_processor_list.pop(0)
            if handler.should_skip(next_output_processor, instance, args, kwargs):
                next_output_processor = None
            else:
                break
        to_wrap = to_wrap.copy()
        to_wrap['output_processor_list'] = next_output_processor_list
        to_wrap['output_processor'] = next_output_processor
    return to_wrap

def has_more_processors(to_wrap) -> bool:
    return len(to_wrap.get('output_processor_list', [])) > 0


from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper
from monocle_apptrace.instrumentation.metamodel.agents.entities.inference import (
    AGENT,
    AGENT_DELEGATION,
    TOOLS,
    AGENT_REQUEST
)
from monocle_apptrace.instrumentation.metamodel.agents.agents_processor import (
    constructor_wrapper,
    handoff_constructor_wrapper,
)

AGENTS_METHODS = [
    # Main agent runner methods
    {
        "package": "agents.run",
        "object": "Runner",
        "method": "run",
        "wrapper_method": atask_wrapper,
        "span_handler": "agents_agent_handler",
        "output_processor": AGENT_REQUEST,
    },
    {
        "package": "agents.run",
        "object": "Runner",
        "method": "run_sync",
        "wrapper_method": task_wrapper,
        "span_handler": "agents_agent_handler",
        "output_processor": AGENT_REQUEST,
    },
    # AgentRunner class methods (internal runner)
    {
        "package": "agents.run",
        "object": "AgentRunner",
        "method": "_run_single_turn",
        "wrapper_method": atask_wrapper,
        "span_handler": "agents_agent_handler",
        "output_processor": AGENT,
    },
    # Function tool decorator - wrap the function_tool function directly
    {
        "package": "agents.tool",
        "object": "FunctionTool",
        "method": "__init__",  # Empty string means wrap the function itself
        "wrapper_method": constructor_wrapper,
        "output_processor": TOOLS,
    },
    {
        "package": "agents.handoffs",
        "object": "Handoff",
        "method": "__init__",  # Empty string means wrap the function itself
        "wrapper_method": handoff_constructor_wrapper,
        "output_processor": AGENT_DELEGATION,
    },
]



import os
import json
import logging
from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.constants import (
    AGENT_PREFIX_KEY,
    INFERENCE_AGENT_DELEGATION,
    INFERENCE_TURN_END,
    INFERENCE_TOOL_CALL,
)
from monocle_apptrace.instrumentation.common.utils import (
    Option,
    get_json_dumps,
    try_option,
)

from monocle_apptrace.instrumentation.metamodel.finish_types import map_hf_finish_reason_to_finish_type

logger = logging.getLogger(__name__)

def update_input_span_events(kwargs):
    input_text = ""
    print("DEBUG kwargs:", kwargs)
    if "inputs" in kwargs:
        if isinstance(kwargs["inputs"], list):
            input_text = " | ".join(str(i) for i in kwargs["inputs"])
        else:
            input_text = str(kwargs["inputs"])
    elif "messages" in kwargs:
        input_text = json.dumps(kwargs["messages"])
    return {"input": input_text}  # always a dict with 'input'



def update_output_span_events(result):
    try:
        if hasattr(result, "choices") and result.choices:
            output = [c.message for c in result.choices]
            output_str = json.dumps(output)
            return output_str[:200] + "..." if len(output_str) > 200 else output_str
    except Exception as e:
        logger.warning("Error in update_output_span_events: %s", str(e))
    return ""

def extract_messages(kwargs):
    """Extract system and user messages"""
    try:
        messages = []
        if "system" in kwargs and isinstance(kwargs["system"], str):
            messages.append({"system": kwargs["system"]})
        if 'messages' in kwargs and kwargs['messages']:
            for msg in kwargs['messages']:
                if msg.get('content') and msg.get('role'):
                    messages.append({msg['role']: msg['content']})
        return [get_json_dumps(message) for message in messages]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_messages: %s", str(e))
        return []

def extract_assistant_message(arguments):
    """
    Extract the assistant message from a Mistral response or stream chunks.
    Returns a JSON string like {"assistant": "<text>"}.
    """
    try:
        result = arguments.get("result") if isinstance(arguments, dict) else arguments
        if result is None:
            return ""

        # Handle full response
        if hasattr(result, "choices") and result.choices:
            msg_obj = result.choices[0].message
            return get_json_dumps({msg_obj.role: msg_obj.content})

        # Handle streaming: result might be a list of CompletionEvent chunks
        if isinstance(result, list):
            content = []
            for chunk in result:
                if hasattr(chunk, "data") and hasattr(chunk.data, "choices") and chunk.data.choices:
                    choice = chunk.data.choices[0]
                    if hasattr(choice, "delta") and hasattr(choice.delta, "content"):
                        content.append(choice.delta.content or "")
            return get_json_dumps({"assistant": "".join(content)})

        return ""

    except Exception as e:
        logger.warning("Warning in extract_assistant_message: %s", str(e))
        return ""
    
def update_span_from_llm_response(result, include_token_counts=False):
    tokens = {
        "completion_tokens": getattr(result.usage, "completion_tokens", 0),
        "prompt_tokens": getattr(result.usage, "prompt_tokens", 0),
        "total_tokens": getattr(result.usage, "total_tokens", 0),
    } if include_token_counts else {}
    # Add other metadata fields like finish_reason, etc.
    return {**tokens, "inference_sub_type": "turn_end"}


def get_exception_status_code(exc):
    if exc is None:
        return "success"
    code = getattr(exc, "status_code", None)
    if code == 401:
        return "unauthorized"
    elif code == 403:
        return "forbidden"
    elif code == 404:
        return "not_found"
    return "error"

def map_finish_reason_to_finish_type(finish_reason):
    """Map Hugging Face finish_reason to finish_type, similar to OpenAI mapping."""
    return map_hf_finish_reason_to_finish_type(finish_reason)


def agent_inference_type(result):
    """
    Simple agent inference type logic: if message contains AGENT_PREFIX_KEY,
    mark as delegation; otherwise it's a normal turn_end.
    """
    try:
        assistant_msg = extract_assistant_message(result)
        if assistant_msg and AGENT_PREFIX_KEY in assistant_msg:
            return INFERENCE_AGENT_DELEGATION
    except Exception as e:
        logger.warning("Error in agent_inference_type: %s", str(e))
    return INFERENCE_TURN_END


def extract_finish_reason(result):
    try:
        return getattr(result, "finish_reason", None)
    except Exception:
        return None





import logging
import inspect
from typing import Dict, List, Optional, Any
from functools import wraps
from contextlib import contextmanager, asynccontextmanager
from opentelemetry.context import Context
from monocle_apptrace.instrumentation.common.utils import (
    set_scope, remove_scope, http_route_handler, http_async_route_handler
)

logger = logging.getLogger(__name__)

def start_scope(
    scope_name: str, 
    scope_value: Optional[str] = None,
    context:  Optional[Context] = None
) -> object:
    """
    Start a new scope with the given name and optional value. If no value is provided, a random UUID will be generated.
    All the spans, across traces created after this call will have the scope attached until the scope is stopped.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    
    Returns:
        Token: A token representing the attached context for the scope. This token is to be used later to stop the current scope.
    """
    try:
        # Set the scope using existing utility
        token = set_scope(scope_name, scope_value, context)
        return token
    except Exception as e:
        logger.warning(f"Failed to start scope: {e}")
        return None

def stop_scope(
    token: object
) -> None:
    """
    Stop the active scope. All the spans created after this will not have the scope attached.
    
    Args:
        token: The token that was returned when the scope was started.
    
    Returns:
        None
    """
    try:
        # Remove the scope
        remove_scope(token)
    except Exception as e:
        logger.warning(f"Failed to stop scope: {e}")
    return



@contextmanager
def monocle_trace_scope(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Context manager to start and stop a scope. All the spans, across traces created within the encapsulated code will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    token = None
    if scope_name:
        token = start_scope(scope_name, scope_value)
    try:
        yield
    finally:
        stop_scope(token)

@asynccontextmanager
async def amonocle_trace_scope(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Async context manager to start and stop a scope. All the spans, across traces created within the encapsulated code will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    token = start_scope(scope_name, scope_value)
    try:
        yield
    finally:
        stop_scope(token)
    
def monocle_trace_scope_method(
    scope_name: str, 
    scope_value: Optional[str] = None
):
    """
    Decorator to start and stop a scope for a method. All the spans, across traces created in the method will have the scope attached.
    
    Args:
        scope_name: The name of the scope.
        scope_value: Optional value of the scope. If None, a random UUID will be generated.
    """
    def decorator(func):
        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                async with amonocle_trace_scope(
                    scope_name, scope_value
                ):
                    result = await func(*args, **kwargs)
                    return result
            return wrapper
        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                with monocle_trace_scope(
                    scope_name, scope_value
                ):
                    result = func(*args, **kwargs)
                    return result
            return wrapper
    return decorator



from monocle_apptrace.instrumentation.common.wrapper import atask_wrapper, task_wrapper
from monocle_apptrace.instrumentation.metamodel.litellm.entities.inference import INFERENCE

LITELLM_METHODS = [
    {
        "package": "litellm.llms.openai.openai",
        "object": "OpenAIChatCompletion",
        "method": "completion",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    },
    {
        "package": "litellm.llms.azure.azure",
        "object": "AzureChatCompletion",
        "method": "completion",
        "wrapper_method": task_wrapper,
        "output_processor": INFERENCE
    }
]



import logging
from threading import local
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes, get_exception_status_code
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from monocle_apptrace.instrumentation.common.utils import MonocleSpanException
from urllib.parse import unquote
from opentelemetry.context import get_current
from opentelemetry.trace import Span, get_current_span
from opentelemetry.trace.propagation import _SPAN_KEY

logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000

def get_route(args) -> str:
    return args[0]['PATH_INFO'] if 'PATH_INFO' in args[0] else ""

def get_method(args) -> str:
    return args[0]['REQUEST_METHOD'] if 'REQUEST_METHOD' in args[0] else ""

def get_params(args) -> dict:
    params = args[0]['QUERY_STRING'] if 'QUERY_STRING' in args[0] else ""
    return unquote(params)

def get_url(args) -> str:
    url = ""
    if len(args) > 1 or not isinstance(args[0], dict):
        if 'HTTP_HOST' in args[0]:
            url = f"http://{args[0]['HTTP_HOST']}{args[0].get('REQUEST_URI', '')}"

    return url

def get_body(args) -> dict:
    return ""

def extract_response(instance) -> str:
    if hasattr(instance, 'data') and hasattr(instance, 'content_length'):
        response = instance.data[0:max(instance.content_length, MAX_DATA_LENGTH)]
    else:
        response = ""
    return response

def extract_status(arguments) -> str:
    if arguments["exception"] is not None:
        return get_exception_status_code(arguments)
    instance = arguments['instance']
    if hasattr(instance, 'status_code'):
        status = f"{instance.status_code}"
        if status not in HTTP_SUCCESS_CODES:
            error_message = extract_response(instance)
            raise MonocleSpanException(f"error: {status} - {error_message}", status)
    else:
        status = "success"
    return status

def flask_pre_tracing(args):
    headers = dict()
    for key, value in args[0].items():
        if key.startswith("HTTP_"):
            new_key = key[5:].lower().replace("_", "-")
            headers[new_key] = value
    return extract_http_headers(headers)

def flask_post_tracing(token):
    clear_http_scopes(token)

class FlaskSpanHandler(SpanHandler):

    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return flask_pre_tracing(args)
    
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value, token):
        flask_post_tracing(token)

class FlaskResponseSpanHandler(SpanHandler):
    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value):
        try:
            _parent_span_context = get_current()
            if _parent_span_context is not None:
                parent_span: Span = _parent_span_context.get(_SPAN_KEY, None)
                if parent_span is not None:
                    self.hydrate_events(to_wrap, wrapped, instance, args, kwargs, return_value, parent_span=parent_span)
        except Exception as e:
            logger.info(f"Failed to propogate flask response: {e}")
        super().post_tracing(to_wrap, wrapped, instance, args, kwargs, return_value)



from monocle_apptrace.instrumentation.common.constants import AGENT_REQUEST_SPAN_NAME, SPAN_SUBTYPES, SPAN_TYPES
from monocle_apptrace.instrumentation.common.utils import get_error_message
from monocle_apptrace.instrumentation.metamodel.agents import _helper
from monocle_apptrace.instrumentation.common.utils import get_error_message

AGENT = {
    "type": SPAN_TYPES.AGENTIC_INVOCATION,
    "subtype": SPAN_SUBTYPES.ROUTING,
    "attributes": [
        [
            {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments: _helper.AGENTS_AGENT_NAME_KEY,
            },
            {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_agent_name(arguments),
            },
            {
                "_comment": "agent description",
                "attribute": "description",
                "accessor": lambda arguments: _helper.get_agent_description(arguments),
            },
            {
                "_comment": "agent instructions",
                "attribute": "instructions",
                "accessor": lambda arguments: _helper.get_agent_instructions(arguments),
            },
        ]
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is Agent input",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_agent_input(
                        arguments
                    ),
                }
            ],
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "attribute": "error_code",
                    "accessor": lambda arguments: get_error_message(arguments)
                },
                {
                    "_comment": "this is response from Agent",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_agent_response(
                        arguments["result"]
                    ),
            },
            {
                "attribute": "error_code",
                "accessor": lambda arguments: get_error_message(arguments)
                }
            ],
        },
        {
            "name": "metadata",
            "attributes": [
                {
                    "_comment": "this is metadata from Agent response",
                    "accessor": lambda arguments: _helper.update_span_from_agent_response(
                        arguments["result"]
                    ),
                }
            ],
        },
    ],
}

AGENT_REQUEST = {
    "type": SPAN_TYPES.AGENTIC_REQUEST,
    "subtype": SPAN_SUBTYPES.PLANNING,
    "attributes": [
        [
            {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments: _helper.AGENTS_AGENT_NAME_KEY,
            }
        ],
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is Agent input",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_agent_input(
                        arguments
                    ),
                }
            ],
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is response from Agent",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_agent_response(
                        arguments["result"]
                    ),
            },
            {
                "attribute": "error_code",
                "accessor": lambda arguments: get_error_message(arguments)
            }],
        },
    ],
}

TOOLS = {
    "type": SPAN_TYPES.AGENTIC_TOOL_INVOCATION,
    "subtype": SPAN_SUBTYPES.ROUTING,
    "attributes": [
        [
            {
                "_comment": "tool type",
                "attribute": "type",
                "accessor": lambda arguments: _helper.get_tool_type(arguments["span"]),
            },
            {
                "_comment": "name of the tool",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_tool_name(
                    arguments["instance"]
                ),
            },
            {
                "_comment": "tool description",
                "attribute": "description",
                "accessor": lambda arguments: _helper.get_tool_description(
                    arguments["instance"]
                ),
            },
        ],
        [
            {
                "_comment": "name of the agent",
                "attribute": "name",
                "accessor": lambda arguments: _helper.get_source_agent(),
            },
            {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments: _helper.AGENTS_AGENT_NAME_KEY,
            },
        ],
    ],
    "events": [
        {
            "name": "data.input",
            "attributes": [
                {
                    "_comment": "this is Tool input",
                    "attribute": "input",
                    "accessor": lambda arguments: _helper.extract_tool_input(arguments),
                },
            ],
        },
        {
            "name": "data.output",
            "attributes": [
                {
                    "_comment": "this is response from Tool",
                    "attribute": "response",
                    "accessor": lambda arguments: _helper.extract_tool_response(
                        arguments["result"]
                    ),
                },
                {
                "attribute": "error_code",
                "accessor": lambda arguments: get_error_message(arguments)
                }
            ],
        },
    ],
}

AGENT_DELEGATION = {
    "type": SPAN_TYPES.AGENTIC_DELEGATION,
    "subtype": SPAN_SUBTYPES.ROUTING,
    "attributes": [
        [
            {
                "_comment": "agent type",
                "attribute": "type",
                "accessor": lambda arguments: _helper.AGENTS_AGENT_NAME_KEY,
            },
            {
                "_comment": "name of the source agent",
                "attribute": "from_agent",
                "accessor": lambda arguments: _helper.get_source_agent(),
            },
            {
                "_comment": "name of the target agent",
                "attribute": "to_agent",
                "accessor": lambda arguments: _helper.extract_handoff_target(arguments),
            },
        ]
    ],
}



from opentelemetry.context import get_value
from monocle_apptrace.instrumentation.common.utils import (
    resolve_from_alias,
    get_json_dumps,
)
import logging

logger = logging.getLogger(__name__)

DELEGATION_NAME_PREFIX = "transfer_to_"
ROOT_AGENT_NAME = "AgentsSDK"
AGENTS_AGENT_NAME_KEY = "agent.openai_agents"


def extract_agent_response(response):
    """Extract the final output from an Agents SDK RunResult."""
    try:
        if response is not None and hasattr(response, "final_output"):
            return str(response.final_output)
        elif (
            response is not None
            and hasattr(response, "new_items")
            and response.new_items
        ):
            # Try to extract from new_items if final_output is not available
            last_item = response.new_items[-1]
            if hasattr(last_item, "content"):
                return str(last_item.content)
            elif hasattr(last_item, "text"):
                return str(last_item.text)
        elif (
            response is not None
            and hasattr(response, "next_step")
            and response.next_step
            and hasattr(response.next_step, "output")
            and response.next_step.output
        ):
            return str(response.next_step.output)
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_agent_response: %s", str(e))
    return ""


def extract_agent_input(arguments):
    """Extract the input provided to the agent."""
    try:
        # For Runner.run and Runner.run_sync, the structure is:
        # Runner.run(starting_agent, input, **kwargs)
        # So args[0] = starting_agent, args[1] = input
        if len(arguments["args"]) > 1:
            input_data = arguments["args"][1]
            if isinstance(input_data, str):
                return [input_data]
            elif isinstance(input_data, list):
                # Handle list of input items
                return input_data

        # Fallback to kwargs
        if "original_input" in arguments["kwargs"]:
            input_data = arguments["kwargs"]["original_input"]
            if isinstance(input_data, str):
                return [input_data]
            elif isinstance(input_data, list):
                return input_data
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_agent_input: %s", str(e))
    return []


def get_agent_name(arguments) -> str:
    """Get the name of an agent."""
    instance = None
    # For Runner methods, the agent is passed as an argument, not instance
    if arguments["args"] and len(arguments["args"]) > 0:
        instance = arguments["args"][0]
    else:
        instance = arguments["kwargs"].get("agent")
    if instance is None:
        return "Unknown Agent"
    return get_name(instance)


def get_agent_description(arguments) -> str:
    """Get the description of an agent."""
    instance = None
    # For Runner methods, the agent is passed as an argument, not instance
    if arguments["args"] and len(arguments["args"]) > 0:
        instance = arguments["args"][0]
    else:
        instance = arguments["kwargs"].get("agent")
    if instance is None:
        return ""
    return get_description(instance)


def get_agent_instructions(arguments) -> str:
    """Get the instructions of an agent."""
    instance = None
    # For Runner methods, the agent is passed as an argument, not instance
    if arguments["args"] and len(arguments["args"]) > 0:
        instance = arguments["args"][0]
    else:
        instance = arguments["kwargs"].get("agent")
    if instance is None:
        return ""
    if hasattr(instance, "instructions"):
        instructions = instance.instructions
        if isinstance(instructions, str):
            return instructions
        elif callable(instructions):
            # For dynamic instructions, we can't easily extract them without context
            return "Dynamic instructions (function)"
    return ""


def extract_tool_response(result):
    """Extract response from a tool execution."""
    if result is not None:
        if hasattr(result, "output"):
            return str(result.output)
        elif hasattr(result, "content"):
            return str(result.content)
        else:
            return str(result)
    return None


def extract_tool_input(arguments):
    """Extract input arguments passed to a tool."""
    try:
        # For function tools, the input is typically in args
        if len(arguments["args"]) > 1:
            tool_input = arguments["args"][
                1
            ]  # Second argument is usually the JSON string with params
            if isinstance(tool_input, str):
                return [tool_input]
            elif isinstance(tool_input, dict):
                return [get_json_dumps(tool_input)]

        # Fallback to all args
        return [str(arg) for arg in arguments["args"]]
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_tool_input: %s", str(e))
    return []


def get_name(instance):
    """Get the name of an agent or tool instance."""
    if hasattr(instance, "name"):
        return str(instance.name)
    elif hasattr(instance, "__name__"):
        return str(instance.__name__)
    return ""


def get_runner_agent_name(instance) -> str:
    """Get the name of an agent."""
    return get_name(instance)

def get_tool_type(span):
    if (span.attributes.get("is_mcp", False)):
        return "tool.mcp"
    else:
        return "tool.openai_agents"

def get_tool_name(instance) -> str:
    """Get the name of a tool."""
    return get_name(instance)


def is_root_agent_name(instance) -> bool:
    """Check if this is the root agent."""
    return get_name(instance) == ROOT_AGENT_NAME


def get_source_agent() -> str:
    """Get the name of the agent that initiated the request."""
    from_agent = get_value(AGENTS_AGENT_NAME_KEY)
    return from_agent if from_agent is not None else ""


def get_description(instance) -> str:
    """Get the description of an instance."""
    if hasattr(instance, "description"):
        return str(instance.description)
    elif hasattr(instance, "handoff_description"):
        return str(instance.handoff_description)
    elif hasattr(instance, "__doc__") and instance.__doc__:
        return str(instance.__doc__)
    return ""


def get_tool_description(instance) -> str:
    """Get the description of a tool."""
    return get_description(instance)


def extract_handoff_target(arguments):
    """Extract the target agent from a handoff operation."""
    try:
        # Check if this is a handoff by looking at the result
        return arguments.get("result").name
    except Exception as e:
        logger.warning("Warning: Error occurred in extract_handoff_target: %s", str(e))
    return ""


def update_span_from_agent_response(response):
    """Update span with metadata from agent response."""
    meta_dict = {}
    try:
        if response is not None and hasattr(response, "usage"):
            usage = response.usage
            if hasattr(usage, "completion_tokens"):
                meta_dict.update({"completion_tokens": usage.completion_tokens})
            if hasattr(usage, "prompt_tokens"):
                meta_dict.update({"prompt_tokens": usage.prompt_tokens})
            if hasattr(usage, "total_tokens"):
                meta_dict.update({"total_tokens": usage.total_tokens})
    except Exception as e:
        logger.warning(
            "Warning: Error occurred in update_span_from_agent_response: %s", str(e)
        )
    return meta_dict



from monocle_apptrace.instrumentation.common.wrapper import task_wrapper, atask_wrapper, atask_iter_wrapper
from monocle_apptrace.instrumentation.metamodel.adk.entities.agent import (
    AGENT, REQUEST, DELEGATION
)
from monocle_apptrace.instrumentation.metamodel.adk.entities.tool import (
    TOOL
)

ADK_METHODS = [
    {
      "package": "google.adk.agents.base_agent",
      "object": "BaseAgent",
      "method": "run_async",
      "wrapper_method": atask_iter_wrapper,
      "output_processor_list": [DELEGATION, AGENT]
    },
    {
      "package": "google.adk.tools.function_tool",
      "object": "FunctionTool",
      "method": "run_async",
      "wrapper_method": atask_wrapper,
      "output_processor": TOOL,
    },
    {
      "package": "google.adk.runners",
      "object": "Runner",
      "method": "run_async",
      "wrapper_method": atask_iter_wrapper,
      "output_processor": REQUEST,
    }
]


import logging
from threading import local
from unittest import result
from monocle_apptrace.instrumentation.common.utils import extract_http_headers, clear_http_scopes, get_exception_status_code, try_option, Option, \
    MonocleSpanException
from monocle_apptrace.instrumentation.common.span_handler import SpanHandler
from monocle_apptrace.instrumentation.common.constants import HTTP_SUCCESS_CODES
from urllib.parse import unquote, urlparse, ParseResult

logger = logging.getLogger(__name__)
MAX_DATA_LENGTH = 1000
token_data = local()
token_data.current_token = None

def get_url(args) -> str:
    event = args[1]
    host = event.get('headers', {}).get('Host', '')
    stage = event.get('requestContext', {}).get('stage', '')
    path = event.get('path', '')
    query_params = event.get('queryStringParameters', {})
    scheme = 'https' if event.get('headers', {}).get('X-Forwarded-Proto', 'http') == 'https' else 'http'
    url = f"{scheme}://{host}/{stage}{path}"
    if query_params:
        from urllib.parse import urlencode
        url += '?' + urlencode(query_params)
    return url

def get_route(args) -> str:
    event = args[1]
    route = event.get("path") or event.get("requestContext", {}).get("path")
    return route

def get_method(args) -> str:
    event = args[1]
    http_method = event.get("httpMethod") or event.get("requestContext", {}).get("httpMethod")
    return http_method


def get_params(args) -> dict:
    event = args[1]
    question = None
    query_params = event.get('queryStringParameters', {})
    if isinstance(query_params, dict):
        question = query_params.get('question')
    return question

def get_body(args) -> dict:
    event = args[1]
    body = event.get("body")
    return body

def extract_response(result) -> str:
    if isinstance(result, dict) and 'body' in result:
        response = result['body']
        if isinstance(response, bytes):
            response = response.decode('utf-8', errors='ignore')
    else:
        response = ""
    return response


def extract_status(arguments) -> str:
    if arguments["exception"] is not None:
        return get_exception_status_code(arguments)
    result = arguments['result']
    if isinstance(result, dict) and 'statusCode' in result:
        status = f"{result['statusCode']}"
        if status not in HTTP_SUCCESS_CODES:
            error_message = extract_response(result)
            raise MonocleSpanException(f"error: {status} - {error_message}", status)
    else:
        status = "success"
    return status

def lambda_func_pre_tracing(kwargs):
    headers = kwargs['event'].get('headers', {}) if 'event' in kwargs else {}
    return extract_http_headers(headers)


def lambda_func_post_tracing(token):
    clear_http_scopes(token)


class lambdaSpanHandler(SpanHandler):
    def pre_tracing(self, to_wrap, wrapped, instance, args, kwargs):
        return lambda_func_pre_tracing(kwargs)

    def post_tracing(self, to_wrap, wrapped, instance, args, kwargs, return_value,token):
        lambda_func_post_tracing(token)
